#                              Copyright (C) 2023, UChicago Argonne, LLC
# 
#                                         All Rights Reserved
# 
#         Software Name: SONICC (Scale-dependent Omega for Neutrino Initial Condition Codes)
# 
#                 By: Argonne National Laboratory, University of California, Berkeley
# 
#                                         OPEN SOURCE LICENSE
# 
# Redistribution and use in source and binary forms, with or without modification, are permitted
# provided that the following conditions are met:
# 
#   1. Redistributions of source code must retain the above copyright notice, this list of conditions
#      and the following disclaimer.
# 
#   2. Redistributions in binary form must reproduce the above copyright notice, this list of
#      conditions and the following disclaimer in the documentation and/or other materials provided
#      with the distribution.
# 
#   3. Neither the name of the copyright holder nor the names of its contributors may be used to
#      endorse or promote products derived from this software without specific prior written
#      permission.
#
# ***************************************************************************************************
#                                              DISCLAIMER
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# ***************************************************************************************************

from . import IO
import scipy.integrate
import numpy


class Cosmology:
    """
    Class containing Hubble factor and growth function definitions as well as relevant cosmological parameters.
    The function Hratio should conform to whatever choices are made in your simulation code.
    """

    def __init__(
        self, h, omch2, ombh2, omnh2, omk, w, wa, tcmb, neff_massless, neff_massive
    ):

        # Set important parameters
        self.h = h
        self.omega_c = omch2 / h**2
        self.omega_b = ombh2 / h**2
        self.omega_n = omnh2 / h**2
        self.omega_k = omk
        self.w = w
        self.wa = wa
        self.omega_r = 2.471e-5 * (tcmb / 2.725) ** 4 / h**2
        self.f_nu_massless = (7.0 / 8.0) * (4.0 / 11.0) ** (4.0 / 3.0) * neff_massless
        self.f_nu_massive = (7.0 / 8.0) * (4.0 / 11.0) ** (4.0 / 3.0) * neff_massive

        # Compute derived parameters
        self.m_nu = 93.14 * self.h**2 * self.omega_n / neff_massive
        self.omega_cb = self.omega_c + self.omega_b
        self.omega_cbn = self.omega_cb + self.omega_n
        self.omega_l = 1.0 - (
            self.omega_cbn + (1.0 + self.f_nu_massless) * self.omega_r + self.omega_k
        )
        self.gc = self.omega_c / self.omega_cb
        self.gb = self.omega_b / self.omega_cb
        self.fc = self.omega_c / self.omega_cbn
        self.fb = self.omega_b / self.omega_cbn
        self.fn = self.omega_n / self.omega_cbn

        # Make sure this is a massive neutrino cosmology
        IO.CheckError(
            self.omega_n > 0.0, "You have provided a non-massive neutrino cosmology"
        )

        # Print some info to screen
        print("Cosmology class initialized with the following parameters")
        IO.PrintParameter("omega_c", self.omega_c, "{:.6f}")
        IO.PrintParameter("omega_b", self.omega_b, "{:.6f}")
        IO.PrintParameter("omega_n", self.omega_n, "{:.4e}")
        IO.PrintParameter("omega_r", self.omega_r, "{:.4e}")
        IO.PrintParameter("omega_k", self.omega_k, "{:.6f}")
        IO.PrintParameter("omega_l", self.omega_l, "{:.6f}")
        IO.PrintParameter("h", self.h, "{:.6f}")
        IO.PrintParameter("neff_massless", neff_massless, "{:.6f}")
        IO.PrintParameter("neff_massive", neff_massive, "{:.6f}")
        IO.PrintParameter("m_nu", self.m_nu, "{:.6f}")
        IO.PrintParameter("w", self.w, "{:.6f}")
        IO.PrintParameter("wa", self.wa, "{:.6f}")
        print()

    def GrowthFactor(self, z, omega_m, z_primoridal=100000.0, tol=1.0e-5):
        """
        Public function that computes the cosmological growth factor (and its derivative) at redshift z for matter background omega_m.
        Will work with either scalar or array of omega_m values. The derivative of the growth factor is returned as a*H(a)*(dD/da).
        """

        if isinstance(omega_m, numpy.ndarray):
            D = numpy.zeros(omega_m.shape[0], dtype=numpy.float64)
            dD = numpy.zeros(omega_m.shape[0], dtype=numpy.float64)
            for i in range(omega_m.shape[0]):
                D[i], dD[i] = self.__GrowthFactor(z, omega_m[i], z_primoridal, tol)
        else:
            D, dD = self.__GrowthFactor(z, omega_m, z_primoridal, tol)

        return D, dD

    def __GrowthFactor(self, z, omega_m, z_primoridal, tol):
        """
        Private function that computes the cosmological growth factor (and its derivative) at redshift z for matter background omega_m.
        """

        x1 = 1.0 / (1.0 + z_primoridal)
        x2 = 1.0 / (1.0 + z)
        x = numpy.array([x1, x2])
        y1 = numpy.array([x1, 0.0])

        y2 = scipy.integrate.odeint(
            self.__d_gf, y1, x, rtol=tol, atol=tol, args=(omega_m,)
        )[1]
        dplus = y2[0]
        ddot = y2[1]

        x1 = 1.0 / (1.0 + z_primoridal)
        x2 = 1.0
        x = numpy.array([x1, x2])
        y1 = numpy.array([x1, 0.0])

        y2 = scipy.integrate.odeint(
            self.__d_gf, y1, x, rtol=tol, atol=tol, args=(omega_m,)
        )[1]
        gf = dplus / y2[0]
        gd = ddot / y2[0]

        D = gf
        dD = gd

        return D, dD

    def __d_gf(self, y, a, omega_m):
        """
        Growth factor ODE to be integrated.
        """

        dydx = 0.0 * y
        H = self.Hratio(a)
        dydx[0] = y[1] / a / H
        dydx[1] = -2.0 * y[1] / a + 1.5 * omega_m * y[0] / (H * a**4)

        return dydx

    def Hratio(self, a):
        """
        H(a)/H0 as a function of a.
        """

        om_r = (1.0 + self.f_nu_massless) * self.omega_r / a**4
        om_cb = self.omega_cb / a**3
        om_nm = self.Omega_nu_massive_a(a)
        om_lm = (
            self.omega_l
            * numpy.power(a, -3.0 * (1.0 + self.w + self.wa))
            * numpy.exp(-3.0 * self.wa * (1.0 - a))
        )
        om_k = self.omega_k / a**2

        return numpy.sqrt(om_r + om_cb + om_nm + om_lm + om_k)

    def Omega_nu_massive_a(self, a):
        """
        Massive neutrino energy density as a function of a. Taken as the largest of either omega_n/a**3 or f_nu_massive*omega_r/a^4
        """

        mat = self.omega_n / a**3
        rad = self.f_nu_massive * self.omega_r / a**4

        return (mat >= rad) * mat + (rad > mat) * rad

    def fOmegaM(self, a):
        """
        Get fraction of energy density associated with non-relativisitc matter at the given scale factor.
        """

        om_m = self.omega_cbn / a**3
        om_r = (1.0 + self.f_nu_massless) * self.omega_r / a**4
        om_lm = (
            self.omega_l
            * numpy.power(a, -3.0 * (1.0 + self.w + self.wa))
            * numpy.exp(-3.0 * self.wa * (1.0 - a))
        )
        om_k = self.omega_k / a**2

        f = om_m / (om_m + om_r + om_lm + om_k)

        return f
