#                              Copyright (C) 2023, UChicago Argonne, LLC
# 
#                                         All Rights Reserved
# 
#         Software Name: SONICC (Scale-dependent Omega for Neutrino Initial Condition Codes)
# 
#                 By: Argonne National Laboratory, University of California, Berkeley
# 
#                                         OPEN SOURCE LICENSE
# 
# Redistribution and use in source and binary forms, with or without modification, are permitted
# provided that the following conditions are met:
# 
#   1. Redistributions of source code must retain the above copyright notice, this list of conditions
#      and the following disclaimer.
# 
#   2. Redistributions in binary form must reproduce the above copyright notice, this list of
#      conditions and the following disclaimer in the documentation and/or other materials provided
#      with the distribution.
# 
#   3. Neither the name of the copyright holder nor the names of its contributors may be used to
#      endorse or promote products derived from this software without specific prior written
#      permission.
#
# ***************************************************************************************************
#                                              DISCLAIMER
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# ***************************************************************************************************

import matplotlib

matplotlib.use("PDF")
import pylab
import numpy
import sys
import os

len_header = 96
len_pprint = 16


def PrintWelcomeMessage():
    """
    Print welcome message for main program to screen.
    """

    global len_header

    print()
    print(len_header * "-")
    print("Welcome to SONICC!".upper().center(len_header))
    print(
        "SONICC uses an approximate method to backscale CAMB transfer functions from".upper().center(
            len_header
        )
    )
    print(
        "massive neutrino cosmologies for use in standard simulation codes that do not".upper().center(
            len_header
        )
    )
    print(
        "include radiation and metric perturbations in their forward evolution".upper().center(
            len_header
        )
    )
    print(len_header * "-")
    print()


def PrintParameter(pname, pvalue, pformat):
    """
    Print the name and value of a given parameter to screen.
    """

    s = "{:>" + str(len_pprint) + "} = " + pformat
    print(s.format(pname, pvalue))


def PrintProgressBar(i, n, bar_length=50):
    """
    Prints a progress bar to screen
    """

    tick = float(n) / bar_length

    s_len = int(numpy.round((i + 1) / tick))
    s_per = int(numpy.round(100.0 * float(i + 1) / n))
    s_pro = (s_len * "=").ljust(bar_length + 1) + str(s_per) + "%\r"
    sys.stdout.write(s_pro)
    sys.stdout.flush()

    if (i + 1) == n:
        print("")


def PrintFinishMessage(time_elapsed):
    """
    Print message to screen when the main program finishes. This includes timing (in seconds).
    """

    global len_header

    print()
    print(len_header * "-")
    print("Finished!".upper().center(len_header))
    print("Total runtime: {:8.2f}".format(time_elapsed).upper().center(len_header))
    print(len_header * "-")
    print()


def CreateOutputDirectory(output_path):
    """
    Create output directory if it does not yet exist.
    """

    if not os.path.isdir(output_path):
        os.mkdir(output_path)


def CheckError(condition, message):
    """
    Raise a runtime error with the given message if the condition failed.
    """

    if not condition:
        raise RuntimeError(message)


def CheckWarning(condition, message):
    """
    Print a warning message to screen if the given condition has failed.
    """

    if not condition:
        print("\n" + 8 * ">" + " WARNING: " + message + "\n")

    return not condition


def WriteOmegaGrav(output_file, k, omega_grav, omega_cbn, omega_cb, output_path):
    """
    Write a text file containing the converged omega_grav as a function of k
    """

    fw = open(output_path + "/" + output_file, "w")
    shead = "# omega_cbn = {:.8f}\n".format(omega_cbn)
    shead += "# omega_cb  = {:.8f}\n".format(omega_cb)
    shead += "{:<16s} {:<16s}\n".format("# k [h/Mpc]", "omega_grav")
    fw.write(shead)
    for i in range(k.shape[0]):
        s = "{:<16.4e} {:<16.8f}\n".format(k[i], omega_grav[i])
        fw.write(s)
    fw.close()

    print("Data saved to " + output_path + "/" + output_file)


def WriteTk(output_file, k, Tc, Tb, Tcb, Tn, Tcbn, output_path):
    """
    Writes cdm, baryon, cdm+baryon, neutrino, cdm+baryon+neutrino transfer functions components to file.
    """

    fw = open(output_path + "/" + output_file, "w")
    shead = "{:<16s} {:<16s} {:<16s} {:<16s} {:<16s} {:<16s}\n".format(
        "# k [h/Mpc]", "Tc", "Tb", "Tcb", "Tn", "Tcbn"
    )
    fw.write(shead)
    for i in range(k.shape[0]):
        s = "{:<16.4e} {:<16.4e} {:<16.4e} {:<16.4e} {:<16.4e} {:<16.4e}\n".format(
            k[i], Tc[i], Tb[i], Tcb[i], Tn[i], Tcbn[i]
        )
        fw.write(s)
    fw.close()

    print("Data saved to " + output_path + "/" + output_file)


def PlotOmegaGrav(
    plot_file,
    k,
    converged,
    omega_grav,
    omega_cbn,
    omega_cb,
    Tkf,
    Tkf_camb,
    ENABLE_OMEGA_PLOTS,
    output_path,
):
    """
    Make a plot showing the current choice form omega_grav with all of the converged points highlighed
    as well as the relative difference between the forward-evolved Tkf versus that of CAMB.
    """

    if not ENABLE_OMEGA_PLOTS:
        return

    nc = numpy.sum(converged)
    ac = numpy.sum(converged) == converged.shape[0]

    #
    # Plot parameters
    #

    # Ranges
    xmin = k.min()
    xmax = k.max()
    dom = 0.1 * (omega_cbn - omega_cb)
    ymin1 = omega_cb - dom
    ymax1 = omega_cbn + dom
    ymin2 = -0.025
    ymax2 = 0.025
    eps = 0.001

    # Labels
    xlabel = r"$k\ [h\ {\rm Mpc}^{-1}]$"
    ylabel1 = r"$\Omega_{\rm grav}$"
    ylabel2 = r"$T_{cbn}^{\rm FE}/T_{cbn}^{\rm CAMB}-1$"

    # Line colors
    cm = "r-"
    cc = "go"
    lm = 2.0
    lc = 2.0
    if ac:
        cm = "g-"

    # Plot adjustments
    left = 0.20
    right = 0.98
    bottom = 0.10
    top = 0.97

    #
    # Make the plot
    #

    fig_width_pt = 400.0
    inches_per_pt = 1.0 / 72.27
    golden_mean = (numpy.sqrt(5.0) - 1.0) / 2.0
    fig_width = fig_width_pt * inches_per_pt
    fig_height = fig_width
    fig_size = [fig_width, fig_height]
    params = {
        "backend": "ps",
        "axes.labelsize": 14,
        "font.size": 14,
        "legend.fontsize": 14,
        "xtick.labelsize": 14,
        "ytick.labelsize": 14,
        "text.usetex": True,
        "figure.figsize": fig_size,
    }

    pylab.rcParams.update(params)
    fig = pylab.figure(1)
    pylab.clf()

    ax1 = pylab.subplot(211)

    pylab.hlines([omega_cb, omega_cbn], xmin, xmax, colors="k", linestyles="dotted")
    if nc > 0 and not ac:
        pylab.semilogx(k[converged], omega_grav[converged], cc, lw=lc)
    pylab.semilogx(k, omega_grav, cm, lw=lm)

    pylab.ylabel(ylabel1)

    ax2 = pylab.subplot(212)

    pylab.hlines(0.0, xmin, xmax, colors="k", linestyles="dotted")
    pylab.axhspan(-eps, eps, facecolor="k", alpha=0.25)

    if nc > 0 and not ac:
        pylab.semilogx(
            k[converged], Tkf[converged] / Tkf_camb[converged] - 1, cc, lw=lc
        )
    pylab.semilogx(k, Tkf / Tkf_camb - 1, cm, lw=lm)

    pylab.ylabel(ylabel2)
    pylab.xlabel(xlabel)

    #
    # Adjustments
    #

    ax1.set_xlim((xmin, xmax))
    ax1.set_ylim((ymin1, ymax1))
    ax2.set_xlim((xmin, xmax))
    # ax2.set_ylim((ymin2, ymax2))

    ax1.tick_params(which="both", direction="in")
    ax1.xaxis.set_ticks_position("both")
    ax1.yaxis.set_ticks_position("both")
    ax2.tick_params(which="both", direction="in")
    ax2.xaxis.set_ticks_position("both")
    ax2.yaxis.set_ticks_position("both")

    pylab.setp(ax1.get_xticklabels(), visible=False)

    pylab.subplots_adjust(
        left=left, right=right, bottom=bottom, top=top, hspace=0.0, wspace=0.0
    )

    pylab.savefig(output_path + "/" + plot_file)
    pylab.close()

    print("Plot saved to " + output_path + "/" + plot_file)


def PlotTransferComparison(
    plot_file,
    z,
    k,
    Tc,
    Tb,
    Tn,
    Tcb,
    Tcbn,
    Tc0,
    Tb0,
    Tn0,
    Tcb0,
    Tcbn0,
    Tc_bs,
    Tb_bs,
    Tn_bs,
    Tcb_bs,
    Tcbn_bs,
    output_path,
):
    """
    Make a plot showing the relative difference between the forward-evolved transfer functions with unmodified CAMB ("0") and the
    zi_nu-backscaled transfer functions ("_bs")
    """

    #
    # Plot parameters
    #

    # Ranges
    xmin = k.min()
    xmax = k.max()
    ymin1 = -0.02
    ymax1 = 0.02
    ymin2 = -0.02
    ymax2 = 0.14
    eps = 0.01

    # Labels
    xlabel = r"$k\ [h\ {\rm Mpc}^{-1}]$"
    ylabel1 = r"$T^{\rm fe}/T^{\rm bs}-1$"
    ylabel2 = r"$T^{\rm fe}/T^{\rm CAMB}-1$"
    title = r"$z = {:.4f}$".format(z)

    # Styles
    cc = "b"
    cb = "r"
    cn = "orange"
    ccb = "k"
    lc = "--"
    lb = ":"
    ln = "-"
    lcb = "-"
    lwc = 2.0
    lwb = 2.0
    lwn = 1.0
    lwcb = 3.0

    # Plot adjustments
    left = 0.16
    right = 0.98
    bottom = 0.10
    top = 0.95

    #
    # Make the plot
    #

    fig_width_pt = 400.0
    inches_per_pt = 1.0 / 72.27
    golden_mean = (numpy.sqrt(5.0) - 1.0) / 2.0
    fig_width = fig_width_pt * inches_per_pt
    fig_height = fig_width
    fig_size = [fig_width, fig_height]
    params = {
        "backend": "ps",
        "axes.labelsize": 14,
        "font.size": 14,
        "legend.fontsize": 14,
        "xtick.labelsize": 14,
        "ytick.labelsize": 14,
        "text.usetex": True,
        "figure.figsize": fig_size,
    }

    pylab.rcParams.update(params)
    fig = pylab.figure(1)
    pylab.clf()

    ax1 = pylab.subplot(211)

    pylab.hlines(0.0, xmin, xmax, colors="k", linestyles="dotted")
    pylab.axhspan(-eps, eps, facecolor="k", alpha=0.25)

    pylab.semilogx(k, Tcb / Tcb_bs - 1, lcb, color=ccb, lw=lwcb, label=r"$T_{cb}$")
    pylab.semilogx(k, Tn / Tn_bs - 1, ln, color=cn, lw=lwn, label=r"$T_\nu$")
    pylab.semilogx(k, Tc / Tc_bs - 1, lc, color=cc, lw=lwc, label=r"$T_c$")
    pylab.semilogx(k, Tb / Tb_bs - 1, lb, color=cb, lw=lwb, label=r"$T_b$")

    lg = pylab.legend(loc="upper left", ncol=4, fancybox=True)
    lg.draw_frame(False)

    pylab.title(title)
    pylab.ylabel(ylabel1)

    ax2 = pylab.subplot(212)

    pylab.hlines(0.0, xmin, xmax, colors="k", linestyles="dotted")
    pylab.axhspan(-eps, eps, facecolor="k", alpha=0.25)

    pylab.semilogx(k, Tcb / Tcb0 - 1, lcb, color=ccb, lw=lwcb)
    pylab.semilogx(k, Tn / Tn0 - 1, ln, color=cn, lw=lwn)
    pylab.semilogx(k, Tc / Tc0 - 1, lc, color=cc, lw=lwc)
    pylab.semilogx(k, Tb / Tb0 - 1, lb, color=cb, lw=lwb)

    pylab.xlabel(xlabel)
    pylab.ylabel(ylabel2)

    #
    # Adjustments
    #

    ax1.set_xlim((xmin, xmax))
    ax1.set_ylim((ymin1, ymax1))
    ax2.set_xlim((xmin, xmax))
    ax2.set_ylim((ymin2, ymax2))

    ax1.tick_params(which="both", direction="in")
    ax1.xaxis.set_ticks_position("both")
    ax1.yaxis.set_ticks_position("both")
    ax2.tick_params(which="both", direction="in")
    ax2.xaxis.set_ticks_position("both")
    ax2.yaxis.set_ticks_position("both")

    pylab.setp(ax1.get_xticklabels(), visible=False)

    pylab.subplots_adjust(
        left=left, right=right, bottom=bottom, top=top, hspace=0.0, wspace=0.0
    )

    pylab.savefig(output_path + "/" + plot_file.format(z))
    pylab.close()

    return "Plot saved to " + output_path + "/" + plot_file.format(z)
