#                              Copyright (C) 2023, UChicago Argonne, LLC
# 
#                                         All Rights Reserved
# 
#         Software Name: SONICC (Scale-dependent Omega for Neutrino Initial Condition Codes)
# 
#                 By: Argonne National Laboratory, University of California, Berkeley
# 
#                                         OPEN SOURCE LICENSE
# 
# Redistribution and use in source and binary forms, with or without modification, are permitted
# provided that the following conditions are met:
# 
#   1. Redistributions of source code must retain the above copyright notice, this list of conditions
#      and the following disclaimer.
# 
#   2. Redistributions in binary form must reproduce the above copyright notice, this list of
#      conditions and the following disclaimer in the documentation and/or other materials provided
#      with the distribution.
# 
#   3. Neither the name of the copyright holder nor the names of its contributors may be used to
#      endorse or promote products derived from this software without specific prior written
#      permission.
#
# ***************************************************************************************************
#                                              DISCLAIMER
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# ***************************************************************************************************

from . import IO
import numpy


class Solver:
    """
    Class containing main solver that computes omega_grav as well as the corresponding backscaled density
    and velocity transfer functions at the neutrino and cold matter starts.
    """

    # Maximum number of iterations to use in the computation of omega_grav
    max_iterations = 16

    # Convergence tolerance for matching CAMB at zf as well as for checking monotonic behavior of omega_grav
    tolerance = 1.0e-6

    # Number of points between omega_cb and omega_cbn to use for sampling the growth factor
    ns_growth = 128

    # Name of file that will store converged omega_grav
    output_file_omega_grav = "omegaGrav.dat"

    # Name of text files storing backscaled density and velocity transfer functions at zi
    output_tk_nu = "delta_transfer_znu.dat"
    output_vk_nu = "theta_transfer_znu.dat"
    output_tk_cb = "delta_transfer_zcb.dat"
    output_vk_cb = "theta_transfer_zcb.dat"

    # Name of text file storing the density transfer function at zf
    output_tk_zf = "delta_transfer_zfin.dat"

    # Name of plot file that will show convergence of omega_grav for each iteration
    output_plot_omega_grav = "omegaGrav{:0>2d}.pdf"

    # Name of plot file that will compare the forward-evolved transfer functions with CAMB
    output_plot_camb = "compare_transfer_z{:.4f}.pdf"

    def __init__(
        self, zcbi, zi, zfd, zf, ns, trans, cosmo, output_path, ENABLE_OMEGA_PLOTS
    ):

        # Set important parameters
        self.zi = zi
        self.zfd = zfd
        self.zf = zf
        self.zcbi = zcbi
        self.ns = ns
        self.trans = trans
        self.cosmo = cosmo
        self.output_path = output_path
        self.ENABLE_OMEGA_PLOTS = ENABLE_OMEGA_PLOTS

        # Store some useful derived values
        self.zm = self.zi - self.zfd
        self.zp = self.zi + self.zfd
        self.zcbim = self.zcbi - self.zfd
        self.zcbip = self.zcbi + self.zfd
        self.ai = 1.0 / (1.0 + self.zi)
        self.acbi = 1.0 / (1.0 + self.zcbi)
        self.af = 1.0 / (1.0 + self.zf)
        self.Hi = cosmo.Hratio(self.ai)
        self.Hcbi = cosmo.Hratio(self.acbi)

        # Instantiate main results as None
        self.omega_grav = None
        self.Tki_bs = None
        self.Vki_bs = None
        self.Tkcbi_bs = None
        self.Vkcbi_bs = None

        # Determine steps taken in the scale factor
        self.asteps = self.__DetermineSteps()

        # Print some info to screen
        print("Solver class initialized with the following parameters")
        IO.PrintParameter("acbi", self.acbi, "{:.6f}")
        IO.PrintParameter("ai", self.ai, "{:.6f}")
        IO.PrintParameter("af", self.af, "{:.6f}")
        IO.PrintParameter("omega_cbn", self.cosmo.omega_cbn, "{:.6f}")
        IO.PrintParameter("omega_cb", self.cosmo.omega_cb, "{:.6f}")
        IO.PrintParameter("num_steps", self.ns, "{:d}")
        print()

    def BackScale(self):
        """
        Backscale the CAMB transfer function at zf to zi_nu using an approximate method that iteratively solves
        for the scale-dependent effective matter background yielding self-consistent growth between zi_nu and zf.
        """

        # Store the initial and final CAMB transfer functions
        self.Tki = self.trans.GetTransferFunctions(self.zi)
        self.Tkf = self.trans.GetTransferFunctions(self.zf)

        # Store common k axis and the CAMB total matter transfer function at zf
        self.k = self.Tki[self.trans.Tk_k]
        self.Tkf_cbn_camb = self.Tkf[self.trans.Tk_cbn]

        # Store the transfer functions needed for the velocity finite difference
        self.Tkm = self.trans.GetTransferFunctions(self.zm)
        self.Tkp = self.trans.GetTransferFunctions(self.zp)

        # Initialize array containing converged total matter transfer function at zf
        self.Tkf_cbn = numpy.zeros(self.k.shape[0], dtype=self.k.dtype)

        # Initialize array containing converged omega_grav values. Start with omega_cbn on the first iteration
        self.omega_grav = numpy.zeros(self.k.shape[0], dtype=self.k.dtype)

        # Initialize array denoting which k values have converged on omega_grav
        self.converged = numpy.zeros(self.k.shape[0], dtype="bool")

        # Sample the growth factor at different redshifts for various omega_grav values (speeds up doing this every iteration)
        omega_samp = numpy.linspace(
            self.cosmo.omega_cb, self.cosmo.omega_cbn, self.ns_growth
        )
        Di_samp, dDi_samp = self.cosmo.GrowthFactor(self.zi, omega_samp)
        Df_samp, dDf_samp = self.cosmo.GrowthFactor(self.zf, omega_samp)
        Dm_samp, dDm_samp = self.cosmo.GrowthFactor(self.zm, omega_samp)
        Dp_samp, dDp_samp = self.cosmo.GrowthFactor(self.zp, omega_samp)

        #
        # Iteratively solve for omega_grav as a function of k. Start with omega_cbn on the first iteration which
        # will allow large scales to converge first. On the second iteration try omega_cb which will allow small
        # scales to converge. On subsequent iterations we interpolate between the minimum and maximum values that
        # have been tried for each remaining unconverged scale.
        #

        IO.CheckError(
            self.max_iterations > 2,
            "The maximum number of solver iterations is too small",
        )
        for niter in range(self.max_iterations):

            # This will only be set True if we are correcting the converged noisy solution
            use_monotonic = False

            # Print progress to screen
            if numpy.sum(self.converged) == self.k.shape[0]:
                check = self.__CheckSolution()
                if check:
                    print(
                        "Fully converged with maximum deviation {:.4e}".format(
                            numpy.max(numpy.abs(target))
                        )
                    )
                    break
                else:
                    print(
                        "Converged but non-monotonic, cleaning up noisy result and trying again"
                    )
                    self.__SmoothSolution()
                    use_monotonic = True
            else:
                print(
                    "Forward evolving iteration "
                    + str(niter)
                    + " with fraction converged: "
                    + str(numpy.sum(self.converged) / (1.0 * self.converged.shape[0]))
                )

            # Use omega_cbn on first iteration and omega_cb on second iteration. Afterwards use a linear root find to converge
            inds_nc = numpy.where(self.converged == False)
            if niter == 0:
                self.omega_grav[inds_nc] = self.cosmo.omega_cbn
                omega_max = self.omega_grav.copy()
            elif niter == 1:
                self.omega_grav[inds_nc] = self.cosmo.omega_cb
                omega_min = self.omega_grav.copy()
            elif not use_monotonic:
                self.__OmegaGravRootFind(
                    inds_nc, omega_min, omega_max, target_min, target_max
                )

            # Compute growth factor at each k by interpolating previous samples
            Di, dDi = self.__InterpolateGrowths(
                self.omega_grav, omega_samp, Di_samp, dDi_samp
            )
            Df, dDf = self.__InterpolateGrowths(
                self.omega_grav, omega_samp, Df_samp, dDf_samp
            )
            Dp, dDp = self.__InterpolateGrowths(
                self.omega_grav, omega_samp, Dp_samp, dDp_samp
            )
            Dm, dDm = self.__InterpolateGrowths(
                self.omega_grav, omega_samp, Dm_samp, dDm_samp
            )

            # Compute the initial transfer function backscaled from zi using the scale-dependent growth
            Tki_bs = self.trans.BackScaledTransferFunctions(self.Tki, self.Tkf, Di, Df)

            # Compute the corresponding velocity transfer functions
            Tkm_bs = self.trans.BackScaledTransferFunctions(self.Tkm, self.Tkf, Dm, Df)
            Tkp_bs = self.trans.BackScaledTransferFunctions(self.Tkp, self.Tkf, Dp, Df)
            Vki_bs = self.trans.ComputeVelocityTransferFunctions(
                self.zi, self.Hi, self.zp, self.zm, Tkp_bs, Tkm_bs
            )

            # Set the initial conditions to be evolved forward from zi to zf (only pull out unconverged scales)
            norm = Tki_bs[self.trans.Tk_cbn][inds_nc]
            Dc = Tki_bs[self.trans.Tk_c][inds_nc] / norm
            Db = Tki_bs[self.trans.Tk_b][inds_nc] / norm
            thetac = self.ai * Vki_bs[self.trans.Tk_c][inds_nc] / norm
            thetab = self.ai * Vki_bs[self.trans.Tk_b][inds_nc] / norm

            #
            # Forward evolve now using log steps in scale factor
            #

            # Set the ticker noting that we will be making steps in log(a)
            lna_ini = numpy.log(self.asteps[0])

            for nstep in range(1, self.asteps.shape[0]):

                # Determine the substep size
                lna_fin = numpy.log(self.asteps[nstep])
                dlna = lna_fin - lna_ini
                dlna2 = dlna / 2

                # Determine closest CAMB sample points that will be used to interpolate Tnu/Tcb
                self.trans.SetFnuInterpolationPoints(
                    self.asteps[nstep - 1], self.asteps[nstep], inds_nc
                )

                # Integrate forward to the next step
                lna, Dc, Db, thetac, thetab = self.__ForwardEvolve(
                    Dc, Db, thetac, thetab, lna_ini, lna_fin
                )

                # Reset the ticker
                lna_ini = lna

                # Update progress bar to screen
                IO.PrintProgressBar(nstep, self.asteps.shape[0])

            # Consistency check
            a = numpy.exp(lna)
            warning = IO.CheckWarning(
                numpy.abs(1.0 - a / self.af) < 1.0e-8,
                "Solver did not integrate to af: {:.4f} {:.4f}".format(a, self.af),
            )

            # Extract the final total matter transfer function at zfin
            Tk_c = Dc * norm
            Tk_b = Db * norm
            Tk_cb = self.cosmo.gc * Tk_c + self.cosmo.gb * Tk_b
            Tk_n = Tk_cb * self.trans.FnuInterpolation(a)
            Tk_cbn = self.cosmo.fc * Tk_c + self.cosmo.fb * Tk_b + self.cosmo.fn * Tk_n

            # Combine back with previously converged results
            self.Tkf_cbn[inds_nc] = Tk_cbn

            # In general, we find a bias on large scales which we subtract off later to make the converged omega_grav smoother
            if niter == 0:
                bias = self.Tkf_cbn[0] / self.Tkf_cbn_camb[0] - 1.0

            # Identify any newly converged scales
            target = self.Tkf_cbn / self.Tkf_cbn_camb - 1.0 - bias
            if niter == 0:
                self.__UpdateConvergedLimits(target)
                target_max = target.copy()
            elif niter == 1:
                self.__UpdateConvergedLimits(target, SMALL_SCALES=True)
                target_min = target.copy()
            elif not use_monotonic:
                omega_min, omega_max, target_min, target_max = self.__UpdateConverged(
                    target, omega_min, omega_max, target_min, target_max
                )
            else:
                self.converged[inds_nc] = True

            # Create plot showing the convergence of omega_grav
            IO.PlotOmegaGrav(
                self.output_plot_omega_grav.format(niter),
                self.k,
                self.converged,
                self.omega_grav,
                self.cosmo.omega_cbn,
                self.cosmo.omega_cb,
                self.Tkf_cbn,
                self.Tkf_cbn_camb,
                ENABLE_OMEGA_PLOTS=self.ENABLE_OMEGA_PLOTS,
                output_path=self.output_path,
            )

        # Write out converged omega_grav(k) to a text file
        IO.WriteOmegaGrav(
            self.output_file_omega_grav,
            self.k,
            self.omega_grav,
            self.cosmo.omega_cbn,
            self.cosmo.omega_cb,
            output_path=self.output_path,
        )

        # Compute initial density and velocity transfer functions
        Di, dDi = self.cosmo.GrowthFactor(self.zi, self.omega_grav)
        Df, dDf = self.cosmo.GrowthFactor(self.zf, self.omega_grav)
        self.Tki_bs = self.trans.BackScaledTransferFunctions(self.Tki, self.Tkf, Di, Df)
        Dm, dDm = self.cosmo.GrowthFactor(self.zm, self.omega_grav)
        Dp, dDp = self.cosmo.GrowthFactor(self.zp, self.omega_grav)
        Tkm_bs = self.trans.BackScaledTransferFunctions(self.Tkm, self.Tkf, Dm, Df)
        Tkp_bs = self.trans.BackScaledTransferFunctions(self.Tkp, self.Tkf, Dp, Df)
        self.Vki_bs = self.trans.ComputeVelocityTransferFunctions(
            self.zi, self.Hi, self.zp, self.zm, Tkp_bs, Tkm_bs
        )

        # Write the results to separate text files
        IO.WriteTk(
            self.output_tk_nu,
            self.Tki_bs[self.trans.Tk_k],
            self.Tki_bs[self.trans.Tk_c],
            self.Tki_bs[self.trans.Tk_b],
            self.Tki_bs[self.trans.Tk_cb],
            self.Tki_bs[self.trans.Tk_n],
            self.Tki_bs[self.trans.Tk_cbn],
            output_path=self.output_path,
        )
        IO.WriteTk(
            self.output_vk_nu,
            self.Vki_bs[self.trans.Tk_k],
            self.Vki_bs[self.trans.Tk_c],
            self.Vki_bs[self.trans.Tk_b],
            self.Vki_bs[self.trans.Tk_cb],
            self.Vki_bs[self.trans.Tk_n],
            self.Vki_bs[self.trans.Tk_cbn],
            output_path=self.output_path,
        )
        IO.WriteTk(
            self.output_tk_zf,
            self.Tkf[self.trans.Tk_k],
            self.Tkf[self.trans.Tk_c],
            self.Tkf[self.trans.Tk_b],
            self.Tkf[self.trans.Tk_cb],
            self.Tkf[self.trans.Tk_n],
            self.Tkf[self.trans.Tk_cbn],
            output_path=self.output_path,
        )

    def BackScaleCB(self):
        """
        Use this function to further backscale the cb in isolation from the backscaled zi_nu transfer functions to zi_cb.
        """

        IO.CheckError(
            self.Tki_bs is not None,
            "Trying to run BackScaleCB without first backscaling to zi_nu",
        )

        # Set the final transfer function which is simply our backscaled results at zi_nu but with the neutrino component removed
        Tkf = self.trans.RemoveNeutrinos(self.Tki_bs)

        # Set the initial transfer function which is the CAMB transfer function backscaled to zi_cb from zi_nu using the
        # scale-independent growth with omega_cbn
        Di, dDi = self.cosmo.GrowthFactor(self.zcbi, self.cosmo.omega_cbn)
        Df, dDf = self.cosmo.GrowthFactor(self.zi, self.cosmo.omega_cbn)
        Tki = self.trans.RemoveNeutrinos(self.trans.GetTransferFunctions(self.zcbi))
        self.Tkcbi_bs = self.trans.BackScaledTransferFunctions(Tki, Tkf, Di, Df)

        # Do the same thing with zi_cb +/- zfd
        Dp, dDp = self.cosmo.GrowthFactor(self.zcbip, self.cosmo.omega_cbn)
        Dm, dDm = self.cosmo.GrowthFactor(self.zcbim, self.cosmo.omega_cbn)
        Tkp = self.trans.RemoveNeutrinos(self.trans.GetTransferFunctions(self.zcbip))
        Tkp = self.trans.BackScaledTransferFunctions(Tkp, Tkf, Dp, Df)
        Tkm = self.trans.RemoveNeutrinos(self.trans.GetTransferFunctions(self.zcbim))
        Tkm = self.trans.BackScaledTransferFunctions(Tkm, Tkf, Dm, Df)

        # Compute the corresponding velocity transfer function at zi_cb
        self.Vkcbi_bs = self.trans.ComputeVelocityTransferFunctions(
            self.zcbi, self.Hcbi, self.zcbip, self.zcbim, Tkp, Tkm
        )

        # Rescale the total velocity transfer function so that it is backscaled as (dDi/dDf) from zi_nu in a scale-independent manner
        # Need to preserve the individual velocities of Tvc and Tvb relative to the total Tvcb
        Fc = self.Vkcbi_bs[self.trans.Tk_c] / self.Vkcbi_bs[self.trans.Tk_cb]
        Fb = self.Vkcbi_bs[self.trans.Tk_b] / self.Vkcbi_bs[self.trans.Tk_cb]
        self.Vkcbi_bs[self.trans.Tk_cb] = (dDi / dDf) * self.Vki_bs[self.trans.Tk_cb]
        self.Vkcbi_bs[self.trans.Tk_cbn] = self.Vkcbi_bs[self.trans.Tk_cb].copy()
        self.Vkcbi_bs[self.trans.Tk_c] = Fc * self.Vkcbi_bs[self.trans.Tk_cb]
        self.Vkcbi_bs[self.trans.Tk_b] = Fb * self.Vkcbi_bs[self.trans.Tk_cb]

        # Write the results to separate text files
        IO.WriteTk(
            self.output_tk_cb,
            self.Tkcbi_bs[self.trans.Tk_k],
            self.Tkcbi_bs[self.trans.Tk_c],
            self.Tkcbi_bs[self.trans.Tk_b],
            self.Tkcbi_bs[self.trans.Tk_cb],
            self.Tkcbi_bs[self.trans.Tk_n],
            self.Tkcbi_bs[self.trans.Tk_cbn],
            output_path=self.output_path,
        )
        IO.WriteTk(
            self.output_vk_cb,
            self.Vkcbi_bs[self.trans.Tk_k],
            self.Vkcbi_bs[self.trans.Tk_c],
            self.Vkcbi_bs[self.trans.Tk_b],
            self.Vkcbi_bs[self.trans.Tk_cb],
            self.Vkcbi_bs[self.trans.Tk_n],
            self.Vkcbi_bs[self.trans.Tk_cbn],
            output_path=self.output_path,
        )

    def ForwardScale(self):
        """
        As a final check we take the backscaled transfer functions and forward scale them to the final redshift. Along the way
        we make plots comparing the relative difference with respect to the unmodifed CAMB transfer functions.
        """

        #
        # Forward scale cb in isolation from zi_cb to zi_nu if applicable
        #

        if self.zcbi > self.zi:

            print("Forward evolving from zi_cb to zi_nu")

            # Set the initial conditions to be evolved from zi_cb to zi_nu
            norm = self.Tkcbi_bs[self.trans.Tk_cbn]
            Dc = self.Tkcbi_bs[self.trans.Tk_c] / norm
            Db = self.Tkcbi_bs[self.trans.Tk_b] / norm
            thetac = self.acbi * self.Vkcbi_bs[self.trans.Tk_c] / norm
            thetab = self.acbi * self.Vkcbi_bs[self.trans.Tk_b] / norm

            # Use the same number of steps to go from zi_cb to zi_nu (probably more than necessary)
            asteps = numpy.linspace(self.acbi, self.ai, self.ns + 1)

            # Set the ticker noting that we will be making steps in log(a)
            lna_ini = numpy.log(asteps[0])

            for nstep in range(1, asteps.shape[0]):

                # Determine the substep size
                lna_fin = numpy.log(asteps[nstep])
                dlna = lna_fin - lna_ini
                dlna2 = dlna / 2

                # Integrate forward to the next step with neutrinos omitted
                lna, Dc, Db, thetac, thetab = self.__ForwardEvolve(
                    Dc, Db, thetac, thetab, lna_ini, lna_fin, ONLY_CB=True
                )

                # Reset the ticker
                lna_ini = lna

                # Update progress bar to screen
                IO.PrintProgressBar(nstep, asteps.shape[0])

            # Consistency check
            a = numpy.exp(lna)
            warning = IO.CheckWarning(
                numpy.abs(1.0 - a / self.ai) < 1.0e-8,
                "Solver did not integrate to ai: {:.4f} {:.4f}".format(a, self.ai),
            )

        else:

            # Set the initial conditions to be evolved from zi_nu to zf
            norm = self.Tki_bs[self.trans.Tk_cbn]
            Dc = self.Tki_bs[self.trans.Tk_c] / norm
            Db = self.Tki_bs[self.trans.Tk_b] / norm
            thetac = self.ai * self.Vki_bs[self.trans.Tk_c] / norm
            thetab = self.ai * self.Vki_bs[self.trans.Tk_b] / norm

        #
        # Make a plot comparing the forward-evolved result to the unmodified and zi_nu-backscaled transfer functions
        # (the latter should match if zi_cb==zi_nu and differ slightly on small scales otherwise)
        #

        # Extract the forward-evolved result
        Tk_c = Dc * norm
        Tk_b = Db * norm
        Tk_cb = self.cosmo.gc * Tk_c + self.cosmo.gb * Tk_b
        Tk_n = Tk_cb * self.Tki[self.trans.Tk_n] / self.Tki[self.trans.Tk_cb]
        Tk_cbn = self.cosmo.fc * Tk_c + self.cosmo.fb * Tk_b + self.cosmo.fn * Tk_n

        # Make a plot comparing the forward-evolved result with the unmodified CAMB transfer function
        sout = IO.PlotTransferComparison(
            self.output_plot_camb,
            self.zi,
            self.Tki[self.trans.Tk_k],
            Tk_c,
            Tk_b,
            Tk_n,
            Tk_cb,
            Tk_cbn,
            self.Tki[self.trans.Tk_c],
            self.Tki[self.trans.Tk_b],
            self.Tki[self.trans.Tk_n],
            self.Tki[self.trans.Tk_cb],
            self.Tki[self.trans.Tk_cbn],
            self.Tki_bs[self.trans.Tk_c],
            self.Tki_bs[self.trans.Tk_b],
            self.Tki_bs[self.trans.Tk_n],
            self.Tki_bs[self.trans.Tk_cb],
            self.Tki_bs[self.trans.Tk_cbn],
            output_path=self.output_path,
        )
        print(sout)

        #
        # Forward scale cb along wiht neutrinos from zi_nu to zf
        #

        # Determine steps which will be the same as before but include any more steps used for comparisons
        asteps = self.asteps.copy()
        acomps = 1.0 / (self.trans.redshifts_comp + 1.0)
        asteps = numpy.unique(numpy.append(asteps, acomps))
        isComp = numpy.isin(asteps, acomps, assume_unique=True)

        # Set the ticker noting that we will be making steps in log(a)
        lna_ini = numpy.log(self.asteps[0])

        # Keep track of some things
        Df, dDf = self.cosmo.GrowthFactor(self.zf, self.omega_grav)
        sout0 = numpy.array([], dtype="<U256")

        print("Forward evolving from zi_nu to zf")

        for nstep in range(1, asteps.shape[0]):

            # Determine the substep size
            lna_fin = numpy.log(asteps[nstep])
            dlna = lna_fin - lna_ini
            dlna2 = dlna / 2

            # Determine closest CAMB sample points that will be used to interpolate Tnu/Tcb
            self.trans.SetFnuInterpolationPoints(asteps[nstep - 1], asteps[nstep])

            # Integrate forward to the next step
            lna, Dc, Db, thetac, thetab = self.__ForwardEvolve(
                Dc, Db, thetac, thetab, lna_ini, lna_fin
            )

            # Reset the ticker
            lna_ini = lna

            # Make a comparison plot if we applicable
            if isComp[nstep]:
                aa = asteps[nstep]
                zz = 1.0 / aa - 1.0
                # Extract the forward-evolved result
                Tk_c = Dc * norm
                Tk_b = Db * norm
                Tk_cb = self.cosmo.gc * Tk_c + self.cosmo.gb * Tk_b
                Tk_n = Tk_cb * self.trans.FnuInterpolation(aa)
                Tk_cbn = (
                    self.cosmo.fc * Tk_c + self.cosmo.fb * Tk_b + self.cosmo.fn * Tk_n
                )
                # Extract unmodified CAMB transfer function
                Tk0 = self.trans.GetTransferFunctions(zz)
                # Compute the backscaled prediction at this time
                D0, dD0 = self.cosmo.GrowthFactor(zz, self.omega_grav)
                Tk0_bs = self.trans.BackScaledTransferFunctions(Tk0, self.Tkf, D0, Df)
                # Make a plot comparing the forward-evolved result with the unmodified CAMB transfer function
                sout = IO.PlotTransferComparison(
                    self.output_plot_camb,
                    zz,
                    self.Tkf[self.trans.Tk_k],
                    Tk_c,
                    Tk_b,
                    Tk_n,
                    Tk_cb,
                    Tk_cbn,
                    Tk0[self.trans.Tk_c],
                    Tk0[self.trans.Tk_b],
                    Tk0[self.trans.Tk_n],
                    Tk0[self.trans.Tk_cb],
                    Tk0[self.trans.Tk_cbn],
                    Tk0_bs[self.trans.Tk_c],
                    Tk0_bs[self.trans.Tk_b],
                    Tk0_bs[self.trans.Tk_n],
                    Tk0_bs[self.trans.Tk_cb],
                    Tk0_bs[self.trans.Tk_cbn],
                    output_path=self.output_path,
                )
                sout0 = numpy.append(sout0, sout)

            # Update progress bar to screen
            IO.PrintProgressBar(nstep, asteps.shape[0])

        # Print info to screen
        for i in range(sout0.shape[0]):
            print(sout0[i])

        # Consistency check
        a = numpy.exp(lna)
        warning = IO.CheckWarning(
            numpy.abs(1.0 - a / self.af) < 1.0e-8,
            "Solver did not integrate to af: {:.4f} {:.4f}".format(a, self.af),
        )

    def __DetermineSteps(self):
        """
        Determine the steps taken which are linear in the scale factor.
        """

        return numpy.linspace(self.ai, self.af, self.ns + 1)

    def __OmegaGravRootFind(
        self, inds_nc, omega_min, omega_max, target_min, target_max
    ):
        """
        Use a linear interpolation to determine the root in (omega, target).
        """

        ind_nconv = numpy.where(self.converged == False)
        m = (target_max[ind_nconv] - target_min[ind_nconv]) / (
            omega_max[ind_nconv] - omega_min[ind_nconv]
        )
        b = target_min[ind_nconv]
        d = -b / m

        self.omega_grav[ind_nconv] = omega_min[ind_nconv] + d

        # If the root is outside the bounds (can happen with noise especially if ns_camb is small)
        # then cap omega_grav at the appropriate bound and set this to be converged
        ind_lb = numpy.where(self.omega_grav < self.cosmo.omega_cb)
        ind_ub = numpy.where(self.omega_grav > self.cosmo.omega_cbn)
        self.omega_grav[ind_lb] = self.cosmo.omega_cb
        self.converged[ind_lb] = True
        self.omega_grav[ind_ub] = self.cosmo.omega_cbn
        self.converged[ind_ub] = True

    def __InterpolateGrowths(self, omega_i, omega_p, D_p, dD_p):
        """
        Perform a linear interpolation from (omega_p, [D_p, dD_p]) onto omega_i
        """

        D_i = numpy.interp(omega_i, omega_p, D_p)
        dD_i = numpy.interp(omega_i, omega_p, dD_p)

        return D_i, dD_i

    def __UpdateConvergedLimits(self, target, SMALL_SCALES=False):
        """
        Use the limiting value on large scales (if SMALL_SCALES is False) or small scales (if SMALL_SCALES is True) to determine
        the maximum/minimum k at which departures from that limiting value occur at the level of tolerance.
        """

        if SMALL_SCALES:
            IO.CheckError(
                self.converged[-1] == False, "Rightmost point should not be converged"
            )
            lval = target[-1]
            ival = numpy.where(numpy.abs(target - lval) > self.tolerance)[0][-1]
            self.converged[ival:] = True
        else:
            IO.CheckError(
                self.converged[0] == False, "Leftmost point should not be converged"
            )
            lval = target[0]
            ival = numpy.where(numpy.abs(target - lval) > self.tolerance)[0][0]
            self.converged[:ival] = True

    def __UpdateConverged(self, target, omega_min, omega_max, target_min, target_max):
        """
        Determine any newly converged points and update the min/max bound for those that have not yet converged.
        """

        icon = numpy.where(
            (numpy.abs(target) < self.tolerance) & (self.converged == False)
        )
        self.converged[icon] = True

        sgn_grav = numpy.sign(target).astype(numpy.int16)
        sgn_min = numpy.sign(target_min).astype(numpy.int16)
        sgn_max = numpy.sign(target_max).astype(numpy.int16)

        i_min = numpy.where((self.converged == False) & (sgn_grav == sgn_min))
        i_max = numpy.where((self.converged == False) & (sgn_grav == sgn_max))

        d_min = numpy.abs(target_min[i_min]) - numpy.abs(target[i_min])
        d_max = numpy.abs(target_max[i_max]) - numpy.abs(target[i_max])

        omega_min[i_min] = self.omega_grav[i_min]
        target_min[i_min] = target[i_min]
        omega_max[i_max] = self.omega_grav[i_max]
        target_max[i_max] = target[i_max]

        return omega_min, omega_max, target_min, target_max

    def __ForwardEvolve(self, Dc, Db, thetac, thetab, lna_ini, lna_fin, ONLY_CB=False):
        """
        Forward-evolve from lna_ini to lna_fin using an RK4 integrator.
        """

        # Determine the step size
        dlna = lna_fin - lna_ini
        dlna2 = dlna / 2

        # Start the ticker
        lna = lna_ini

        #
        # Perform the RK4 integration
        #

        k1c, k1b = self.__DeltaEqn(lna, thetac, thetab)
        l1c, l1b = self.__ThetaEqn(lna, Dc, Db, thetac, thetab, ONLY_CB)

        k2c, k2b = self.__DeltaEqn(
            lna + 0.5 * dlna, thetac + 0.5 * dlna * l1c, thetab + 0.5 * dlna * l1b
        )
        l2c, l2b = self.__ThetaEqn(
            lna + 0.5 * dlna,
            Dc + 0.5 * dlna * k1c,
            Db + 0.5 * dlna * k1b,
            thetac + 0.5 * dlna * l1c,
            thetab + 0.5 * dlna * l1b,
            ONLY_CB,
        )

        k3c, k3b = self.__DeltaEqn(
            lna + 0.5 * dlna, thetac + 0.5 * dlna * l2c, thetab + 0.5 * dlna * l2b
        )
        l3c, l3b = self.__ThetaEqn(
            lna + 0.5 * dlna,
            Dc + 0.5 * dlna * k2c,
            Db + 0.5 * dlna * k2b,
            thetac + 0.5 * dlna * l2c,
            thetab + 0.5 * dlna * l2b,
            ONLY_CB,
        )

        k4c, k4b = self.__DeltaEqn(lna + dlna, thetac + dlna * l3c, thetab + dlna * l3b)
        l4c, l4b = self.__ThetaEqn(
            lna + dlna,
            Dc + dlna * k3c,
            Db + dlna * k3b,
            thetac + dlna * l3c,
            thetab + dlna * l3b,
            ONLY_CB,
        )

        lna += dlna
        Dc += (dlna / 6.0) * (k1c + 2 * k2c + 2 * k3c + k4c)
        Db += (dlna / 6.0) * (k1b + 2 * k2b + 2 * k3b + k4b)
        thetac += (dlna / 6.0) * (l1c + 2 * l2c + 2 * l3c + l4c)
        thetab += (dlna / 6.0) * (l1b + 2 * l2b + 2 * l3b + l4b)

        return lna, Dc, Db, thetac, thetab

    def __DeltaEqn(self, lna, thetac, thetab):
        """
        Delta evolution equation:
            d[delta_i]/dlna = -theta_i/aH
        """

        a = numpy.exp(lna)
        H = self.cosmo.Hratio(a)
        C = -1.0 / a / H

        ddeltac = C * thetac
        ddeltab = C * thetab

        return ddeltac, ddeltab

    def __ThetaEqn(self, lna, Dc, Db, thetac, thetab, ONLY_CB):
        """
        Theta evolution equation:
            d[theta_i]/dlna = -(theta_i + 1.5*a*H*Omega_m*delta_m)
        """

        a = numpy.exp(lna)
        H = self.cosmo.Hratio(a)

        A = -1.0
        B = -1.5 * a * H * self.cosmo.fOmegaM(a)
        Dcb = self.cosmo.gc * Dc + self.cosmo.gb * Db
        if ONLY_CB:
            Dm = Dcb
        else:
            Dn = Dcb * self.trans.FnuInterpolation(a)
            Dm = self.cosmo.fc * Dc + self.cosmo.fb * Db + self.cosmo.fn * Dn

        dthetac = A * thetac + B * Dm
        dthetab = A * thetab + B * Dm

        return dthetac, dthetab

    def __CheckSolution(self):
        """
        Consistency check to make sure that omega_grav is monotonically decreasing with increasing k.
        """

        cpass = True
        for i in range(1, self.omega_grav.shape[0] - 1):
            if self.omega_grav[i] > self.omega_grav[i - 1]:
                cpass = False
                break

        return cpass

    def __SmoothSolution(self):
        """
        Interpolates within omega_grav to ensure that this is a monotonically decreasing function with increasing k.
        """

        for i in range(1, self.omega_grav.shape[0] - 1):
            if self.omega_grav[i] > self.omega_grav[i - 1]:
                indxl = i - 1
                indxh = -1
                for j in range(i + 1, self.omega_grav.shape[0]):
                    if self.omega_grav[j] <= self.omega_grav[indxl]:
                        indxh = j
                        break
                warning = IO.CheckWarning(
                    indxh > indxl,
                    "Smoothed solution failed to find an endpoint for interpolation {:d} {:d}".format(
                        indxl, indxh
                    ),
                )
                if warning:
                    break
                x0 = numpy.log(self.k[indxl])
                y0 = self.omega_grav[indxl]
                x1 = numpy.log(self.k[indxh])
                y1 = self.omega_grav[indxh]
                m = (y1 - y0) / (x1 - x0)
                for j in range(indxl + 1, indxh):
                    xi = numpy.log(self.k[j])
                    self.omega_grav[j] = y0 + m * (xi - x0)
                    self.converged[j] = False
