#                              Copyright (C) 2023, UChicago Argonne, LLC
# 
#                                         All Rights Reserved
# 
#         Software Name: SONICC (Scale-dependent Omega for Neutrino Initial Condition Codes)
# 
#                 By: Argonne National Laboratory, University of California, Berkeley
# 
#                                         OPEN SOURCE LICENSE
# 
# Redistribution and use in source and binary forms, with or without modification, are permitted
# provided that the following conditions are met:
# 
#   1. Redistributions of source code must retain the above copyright notice, this list of conditions
#      and the following disclaimer.
# 
#   2. Redistributions in binary form must reproduce the above copyright notice, this list of
#      conditions and the following disclaimer in the documentation and/or other materials provided
#      with the distribution.
# 
#   3. Neither the name of the copyright holder nor the names of its contributors may be used to
#      endorse or promote products derived from this software without specific prior written
#      permission.
#
# ***************************************************************************************************
#                                              DISCLAIMER
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# ***************************************************************************************************

from . import IO
import numpy
import camb


class TransferFunctions:
    """
    Class containing CAMB parameters and transfer functions as well as associated routines.
    """

    # CAMB keys used to grab various transfer function components
    Tk_k = camb.model.transfer_names[camb.model.Transfer_kh - 1]
    Tk_c = camb.model.transfer_names[camb.model.Transfer_cdm - 1]
    Tk_b = camb.model.transfer_names[camb.model.Transfer_b - 1]
    Tk_cb = camb.model.transfer_names[camb.model.Transfer_nonu - 1]
    Tk_n = camb.model.transfer_names[camb.model.Transfer_nu - 1]
    Tk_cbn = camb.model.transfer_names[camb.model.Transfer_tot - 1]

    # Redshift tolerance parameter
    zeps = 1.0e-4

    # CAMB transfer function output
    output_tk_tmp = "camb_transfer_z{:.4f}.dat"

    def __init__(self, camb_params_file, zi_cb, zi_nu, zfd, zf, ns_nu, ns_comp, output_path, output_redshifts):

        # Set important parameters
        self.zi_cb = zi_cb
        self.zi = zi_nu
        self.zfd = zfd
        self.zi_m = self.zi - self.zfd
        self.zi_p = self.zi + self.zfd
        self.zi_cb_m = self.zi_cb - self.zfd
        self.zi_cb_p = self.zi_cb + self.zfd
        self.zf = zf
        self.ns_nu = ns_nu
        self.ns_comp = min(max(2, ns_comp), ns_nu)
        self.enable_output = False
        if len(output_redshifts) > 0: self.enable_output = True
        self.output_path = output_path
        self.output_redshifts = output_redshifts

        # Some sanity checks
        IO.CheckError(
            self.zi_cb >= self.zi,
            "We do not support neutrinos starting earlier than cdm+baryons",
        )
        IO.CheckError(
            self.zi > self.zf, "Initial redshift is smaller than final redshift"
        )
        IO.CheckError(self.zf >= 0.0, "Sorry, we cannot integrate into the future")

        # Read input file to set internal CAMB parameters
        self.camb_params = camb.read_ini(camb_params_file)

        # Reset the desired redshift samples based on the starting and stopping redshifts
        # as well as the intermediate steps needed to sample the relative neutrino growth
        self.redshifts, self.redshifts_comp = self.__SampleRedshifts()
        self.camb_params.set_matter_power(
            redshifts=self.redshifts,
            kmax=self.camb_params.Transfer.kmax,
            k_per_logint=self.camb_params.Transfer.k_per_logint,
            nonlinear=0,
            accurate_massive_neutrino_transfers=self.camb_params.Transfer.accurate_massive_neutrinos,
        )

        # Set flag denoting that CAMB has not yet been run
        self.HaveRunCAMB = False

        # Print some info to screen
        print("Transfer function class initialized with the following parameters")
        IO.PrintParameter("zi_cb", self.zi_cb, "{:.4f}")
        IO.PrintParameter("zi_nu", self.zi, "{:.4f}")
        IO.PrintParameter("zf", self.zf, "{:.4f}")
        IO.PrintParameter("zfd", self.zfd, "{:.4f}")
        IO.PrintParameter("num_redshifts", self.redshifts.shape[0], "{:d}")
        print()

    def RunCAMB(self):
        """
        Computes transfer functions at all desired redshifts with CAMB.
        """

        # Compute transfer fucntions at all redshifts if this has not yet been done
        if self.HaveRunCAMB:
            return
        self.results = camb.get_transfer_functions(self.camb_params)
        self.transfers = self.results.get_matter_transfer_data()

        # Set flag indicating that we have already run CAMB
        self.HaveRunCAMB = True

        # Output CAMB results at select redshfits, if desired
        if self.enable_output: 
            for zz in self.output_redshifts:
                output_tk = self.output_tk_tmp.format(zz)
                Tk = self.GetTransferFunctions(zz)
                IO.WriteTk(     
                    output_tk,
                    Tk[self.Tk_k],
                    Tk[self.Tk_c],
                    Tk[self.Tk_b],
                    Tk[self.Tk_cb],
                    Tk[self.Tk_n],
                    Tk[self.Tk_cbn],
                    output_path=self.output_path,
                )

    def RedshiftIndex(self, z):
        """
        Returns index of transfer functions corresponding to the given redshift. Fails if the
        requested redshift is not sufficiently close.
        """

        dz = numpy.abs(self.redshifts - z)
        indx = numpy.argmin(dz)

        # If this trips then we are requesting a redshift that has not been sampled with CAMB
        IO.CheckError(
            dz[indx] <= self.zeps,
            "The redshift {:.4f} has not been sampled by CAMB".format(z),
        )

        return indx

    def GetTransferFunctions(self, z):
        """
        Returns the transfer functions generated by CAMB at the specified redshift.
        """

        IO.CheckError(
            self.HaveRunCAMB, "Requesting transfer functions before running CAMB"
        )

        indx = self.RedshiftIndex(z)

        Tk = {}
        Tk[self.Tk_k] = self.__Tk(self.Tk_k, indx)
        Tk[self.Tk_c] = self.__Tk(self.Tk_c, indx)
        Tk[self.Tk_b] = self.__Tk(self.Tk_b, indx)
        Tk[self.Tk_cb] = self.__Tk(self.Tk_cb, indx)
        Tk[self.Tk_n] = self.__Tk(self.Tk_n, indx)
        Tk[self.Tk_cbn] = self.__Tk(self.Tk_cbn, indx)

        return Tk

    def RemoveNeutrinos(self, Tk0):
        """
        Sets the Tnu component of the transfer function to 0 and sets Tcbn to be Tcb.
        """

        Tk = {}
        Tk[self.Tk_k] = Tk0[self.Tk_k].copy()
        Tk[self.Tk_c] = Tk0[self.Tk_c].copy()
        Tk[self.Tk_b] = Tk0[self.Tk_b].copy()
        Tk[self.Tk_cb] = Tk0[self.Tk_cb].copy()
        Tk[self.Tk_n] = 0.0 * Tk0[self.Tk_n].copy()
        Tk[self.Tk_cbn] = Tk0[self.Tk_cb].copy()

        return Tk

    def ComputeVelocityTransferFunctions(self, z0, Hz0, z1, z2, Tk1, Tk2):
        """
        Computes the velocity divergence transfer function for each component.
        """

        Vk = {}
        Vk[self.Tk_k] = Tk1[self.Tk_k]
        Vk[self.Tk_c] = self.ComputeVelocityTransferFunction(
            z0, Hz0, z1, z2, Tk1[self.Tk_c], Tk2[self.Tk_c]
        )
        Vk[self.Tk_b] = self.ComputeVelocityTransferFunction(
            z0, Hz0, z1, z2, Tk1[self.Tk_b], Tk2[self.Tk_b]
        )
        Vk[self.Tk_cb] = self.ComputeVelocityTransferFunction(
            z0, Hz0, z1, z2, Tk1[self.Tk_cb], Tk2[self.Tk_cb]
        )
        Vk[self.Tk_n] = self.ComputeVelocityTransferFunction(
            z0, Hz0, z1, z2, Tk1[self.Tk_n], Tk2[self.Tk_n]
        )
        Vk[self.Tk_cbn] = self.ComputeVelocityTransferFunction(
            z0, Hz0, z1, z2, Tk1[self.Tk_cbn], Tk2[self.Tk_cbn]
        )

        return Vk

    def ComputeVelocityTransferFunction(self, z0, Hz0, z1, z2, Tk1, Tk2):
        """
        Computes the (comoving) velocity divergence transfer function using the continuity equation:
            T_theta = -a*H(a)*dT_delta/da
                    = (1+z)*H(z)*dT_delta/dz
        """

        IO.CheckError(z1 != z2, "Finite difference won't work with z1 == z2")

        return (1 + z0) * Hz0 * (Tk2 - Tk1) / (z2 - z1)

    def BackScaledTransferFunctions(self, Tki, Tkf, Di, Df):
        """
        Return the back-scaled transfer function that results from applying the provided
        scale-dependent growth with respect to the total matter transfer function.
        """

        rbs = (Di / Df) * (Tkf[self.Tk_cbn] / Tki[self.Tk_cbn])

        Tk_bs = {}
        Tk_bs[self.Tk_k] = Tki[self.Tk_k]
        Tk_bs[self.Tk_c] = rbs * Tki[self.Tk_c]
        Tk_bs[self.Tk_b] = rbs * Tki[self.Tk_b]
        Tk_bs[self.Tk_cb] = rbs * Tki[self.Tk_cb]
        Tk_bs[self.Tk_n] = rbs * Tki[self.Tk_n]
        Tk_bs[self.Tk_cbn] = rbs * Tki[self.Tk_cbn]

        return Tk_bs

    def SetFnuInterpolationPoints(self, a1, a2, inds_nc=None):
        """
        Determines the indices of the closest sample points that bracket (a1, a2).
        """

        z1 = 1.0 / a1 - 1.0
        z2 = 1.0 / a2 - 1.0

        # Get the closest redshift to z1 without being smaller than z1
        zhigh = self.redshifts[numpy.where(self.redshifts >= z1)]
        self.fnu_z1 = zhigh[numpy.argmin(numpy.abs(z1 - zhigh))]
        self.fnu_a1 = 1.0 / (1.0 + self.fnu_z1)
        self.fnu_indx1 = self.RedshiftIndex(self.fnu_z1)
        self.fnu1 = self.__Tk(self.Tk_n, self.fnu_indx1) / self.__Tk(
            self.Tk_cb, self.fnu_indx1
        )
        if inds_nc is not None:
            self.fnu1 = self.fnu1[inds_nc]

        # Get the closest redshift to z2 without being larger than z2
        zlow = self.redshifts[numpy.where(self.redshifts <= z2)]
        self.fnu_z2 = zlow[numpy.argmin(numpy.abs(z2 - zlow))]
        self.fnu_a2 = 1.0 / (1.0 + self.fnu_z2)
        self.fnu_indx2 = self.RedshiftIndex(self.fnu_z2)
        self.fnu2 = self.__Tk(self.Tk_n, self.fnu_indx2) / self.__Tk(
            self.Tk_cb, self.fnu_indx2
        )
        if inds_nc is not None:
            self.fnu2 = self.fnu2[inds_nc]

    def FnuInterpolation(self, ai, eps=1.0e-4):
        """
        Performs a linear interpolation of Tnu/Tcb based on the bounds set in SetFnuInterpolationPoints
        """

        IO.CheckError(
            (ai >= self.fnu_a1 - eps and ai <= self.fnu_a2 + eps),
            "Fnu interpolation at {:.4f} is outside of bounds: {:.4f} {:.4f}".format(
                ai, self.fnu_a1, self.fnu_a2
            ),
        )

        da = (ai - self.fnu_a1) / (self.fnu_a2 - self.fnu_a1)
        fnui = self.fnu1 + da * (self.fnu2 - self.fnu1)

        return fnui

    def __Tk(self, component, indx):
        """
        Returns the transfer function at the given redshift index for a specific component. If that
        component is either Tcb or Tcbn then we directly compute this as the mass-weighted sum of
        the individual Tc, Tb, and Tn components.
        """

        if component == self.Tk_cb:
            om_c = self.camb_params.omch2
            om_b = self.camb_params.ombh2
            fc = om_c / (om_c + om_b)
            fb = om_b / (om_c + om_b)
            Tcb = fc * self.transfers.transfer_z(
                self.Tk_c, z_index=indx
            ) + fb * self.transfers.transfer_z(self.Tk_b, z_index=indx)
            return Tcb
        elif component == self.Tk_cbn:
            om_c = self.camb_params.omch2
            om_b = self.camb_params.ombh2
            om_n = self.camb_params.omnuh2
            fc = om_c / (om_c + om_b + om_n)
            fb = om_b / (om_c + om_b + om_n)
            fn = om_n / (om_c + om_b + om_n)
            Tcbn = (
                fc * self.transfers.transfer_z(self.Tk_c, z_index=indx)
                + fb * self.transfers.transfer_z(self.Tk_b, z_index=indx)
                + fn * self.transfers.transfer_z(self.Tk_n, z_index=indx)
            )
            return Tcbn
        else:
            return self.transfers.transfer_z(component, z_index=indx)

    def __SampleRedshifts(self):
        """
        Returns a list of redshifts that we need to sample density transfer functions with CAMB.
        These include the starting and stopping redshifts with the possibility that the former is
        different for cb vs nu as well as the intermediate redshifts used to linearly interpolate
        Tnu/Tm from. We additionally need to sample density transfer functions at zi_nu +/- zfd in
        order to compute finite-differenced velocity transfer functions.
        """

        # We will sample Tnu/Tm linearly in the scale factor between ai_nu and af
        scales = numpy.linspace(
            1.0 / (1.0 + self.zi), 1.0 / (1.0 + self.zf), self.ns_nu
        )
        reds = 1.0 / scales - 1.0

        # Determine which of these redshifts will be used for comparison purposes at the end
        reds_comp = reds[
            numpy.round(numpy.linspace(0, scales.shape[0] - 1, self.ns_comp)).astype(
                "int32"
            )
        ]

        # We also need to sample at zi_nu +/- zfd in order to compute velocity transfer functions
        reds = numpy.append(reds, self.zi_p)
        reds = numpy.append(reds, self.zi_m)

        # Finally we need to sample at zi_cb +/- zfd (which may or may not be different than zi_nu)
        reds = numpy.append(reds, self.zi_cb)
        reds = numpy.append(reds, self.zi_cb_m)
        reds = numpy.append(reds, self.zi_cb_p)
    
        # Finally add in any desired output redshifts
        if self.enable_output:
            reds = numpy.append(reds, self.output_redshifts)

        # Remove any potential duplicates and sort in descending order
        reds = numpy.unique(reds)[::-1]

        # Make sure we are not requesing too many redshift samples for CAMB to handle
        IO.CheckError(
            reds.shape[0] <= camb.model.max_transfer_redshifts,
            "Requested {:d} redshift samples which is larger than the {:d} maximum set by CAMB. Decrease ns_camb.".format(
                reds.shape[0], camb.model.max_transfer_redshifts
            ),
        )

        return reds, reds_comp
