#                              Copyright (C) 2023, UChicago Argonne, LLC
# 
#                                         All Rights Reserved
# 
#         Software Name: SONICC (Scale-dependent Omega for Neutrino Initial Condition Codes)
# 
#                 By: Argonne National Laboratory, University of California, Berkeley
# 
#                                         OPEN SOURCE LICENSE
# 
# Redistribution and use in source and binary forms, with or without modification, are permitted
# provided that the following conditions are met:
# 
#   1. Redistributions of source code must retain the above copyright notice, this list of conditions
#      and the following disclaimer.
# 
#   2. Redistributions in binary form must reproduce the above copyright notice, this list of
#      conditions and the following disclaimer in the documentation and/or other materials provided
#      with the distribution.
# 
#   3. Neither the name of the copyright holder nor the names of its contributors may be used to
#      endorse or promote products derived from this software without specific prior written
#      permission.
#
# ***************************************************************************************************
#                                              DISCLAIMER
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
# OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# ***************************************************************************************************

from sonicc.TransferFunctions import TransferFunctions
from sonicc.Cosmology import Cosmology
from sonicc.Solver import Solver
from sonicc.Timer import Timer
import sonicc.IO as IO
from pathlib import Path

# import params.sonicc_params as Parameters
import click
import toml


@click.command()
@click.argument("config_path", type=click.Path(exists=True))
def cli(config_path: str):
    config_dir = Path(config_path).parent
    parameters = toml.load(config_path)
    if Path(parameters["camb_params_file"]).is_absolute():
        camb_file = parameters["camb_params_file"]
    else:
        camb_file = str(config_dir / parameters["camb_params_file"])

    if Path(parameters["output_path"]).is_absolute():
        output_path = parameters["output_path"]
    else:
        output_path = str(config_dir / parameters["output_path"])

    #
    # Start program and timer
    #
    timer = Timer()
    tprog = timer.StartTime()
    IO.PrintWelcomeMessage()
    # IO.CreateOutputDirectory(output_path)
    Path(output_path).mkdir(parents=True, exist_ok=True)

    #
    # Initialize transfer function, cosmology, and solver classes
    #

    trans = TransferFunctions(
        camb_file,
        parameters["zi_cb"],
        parameters["zi_nu"],
        parameters["zfd"],
        parameters["zf"],
        parameters["ns_camb"],
        parameters["ns_compare"],
        output_path=output_path,
        output_redshifts=parameters["camb_redshift_outputs"],
    )

    cosmo = Cosmology(
        trans.camb_params.h,
        trans.camb_params.omch2,
        trans.camb_params.ombh2,
        trans.camb_params.omnuh2,
        trans.camb_params.omk,
        trans.camb_params.DarkEnergy.w,
        trans.camb_params.DarkEnergy.wa,
        trans.camb_params.TCMB,
        trans.camb_params.num_nu_massless,
        trans.camb_params.num_nu_massive,
    )

    solve = Solver(
        parameters["zi_cb"],
        parameters["zi_nu"],
        parameters["zfd"],
        parameters["zf"],
        parameters["ns_solve"],
        trans,
        cosmo,
        output_path=output_path,
        ENABLE_OMEGA_PLOTS=parameters["ENABLE_OMEGA_PLOTS"],
    )

    #
    # Use CAMB to generate transfer functions at starting, stopping, and sampling redshifts
    #

    tcamb = timer.StartTime("Running CAMB to generate transfer functions...")
    trans.RunCAMB()
    timer.PrintElapsedTime(tcamb)

    #
    # Backscale the transfer function from zf to zi_nu using method that iteratively solves for the scale-dependent effective
    # gravitational source term. If cb is started at an earlier time then backscale the cb results using analytic approach
    #

    tsolve = timer.StartTime(
        "Backscaling cbn from z = {:6f} to z = {:6f}...".format(
            parameters["zf"], parameters["zi_nu"]
        )
    )
    solve.BackScale()
    timer.PrintElapsedTime(tsolve)

    #
    # If cb is started at an earlier time then backscale the cb results above to zi_cb
    #

    if parameters["zi_cb"] > parameters["zi_nu"]:
        tsolvecb = timer.StartTime(
            "Backscaling cb in isolation from z = {:6f} to z = {:6f}...".format(
                parameters["zi_nu"], parameters["zi_cb"]
            )
        )
        solve.BackScaleCB()
        timer.PrintElapsedTime(tsolvecb)

    #
    # As a final consistency check we will forward evolve the results to zf with some plots comparing the solution
    # at various times to the unmodified CAMB output
    #

    if parameters["ENABLE_FORWARD_CHECK"]:
        tforward = timer.StartTime(
            "Forward-evolving our final results and making comparison plots with unmodified CAMB outputs"
        )
        solve.ForwardScale()
        timer.PrintElapsedTime(tforward)

    #
    # Finish
    #

    time = timer.ElapsedTime(tprog)
    IO.PrintFinishMessage(time)


if __name__ == "__main__":
    cli()
